;
; Plasmasimulator_v1
; heinz@gfai.de
; 
; PWM-Zeitfunktionsgenerator fr vier LED-Kanle a 4Bit
; Board AT13 mit Atmel ATtiny13
; 
; AVR-Studio 4.15
; 	fertig 12.5.2009 
;	ergnzt 12.5.2010
;	PB4 /STOP statisch 23.04.2013
;	
; berblick
;	Der Plasmasimulator simuliert eine Pulsschweimaschine. Am Eingang PB4 wird das 
;	/STOP - Signal vom Spektralregler Ion angeschlossen. An PB0 bis PB3 knnen vier 
; 	spektral verschiedene LEDs angeschlossen werden.
;	Der Plasmasimulator wiederholt permanent die im include-File gespeicherten 
;	Abtastwerte von vier Kanlen - i.a. einen Schweipuls. 
;	Der include-File wird mit dem Scilab-Programm wav2gif_v6 erzeugt. Dessen Input 
;	sind vorher aufgenommene wav-Files der vier Kanle.
; 	Abweichend von der Schweimaschine ist an PB4
;	ein nichtinvertiertes Stop-Signal zu legen: PB4 = 0: Stop
; 	Dazu am OC-TTL Adapter der Jumper JP2 auf NINV (Brcke Pins 1-2) zu stellen.
;	
; Funktion: 
; - Ausgabe von vier 16-stufigen (4 Bit) Zeitfunktionen mit je 256 Samples
; 	als Pulsweitenmodulation (PWM) fr vier LEDs an PB0-3
; - PWM-Grundtakt = Interrupt von 6,25 s, 
; - pro Wert = 16 Stufen = 16 * 6,25 s = 100 s
; - Zeitfunktionen werden periodisch wiederholt
; - sie sind als Filename.inc mit ".include *.inc" 
;   am Programmende einzubinden
; - an PB4 wacht ein HL-Flankeninterrupt und startet den Ablauf neu
;
; Prozessor: ATMEL ATtiny13V mit 5V und 9,6MHz interner RC-Osz.
.include "tn13def.inc"	; Prozessordaten
; 
; 
; Pinout 8-PDIP/SOIC: 	
; 				 --------
; 	/RESET	PB5	|1	`	8| VCC
; 			PB3	|2		7| PB2	SCK
; 			PB4	|3		6| PB1	MISO
; 			GND	|4		5| PB0	MOSI
; 	 			 --------
; Portpins:
;
; 	Port	Register	Wellenlnge
;   -----------------------------------------
; 	PB0:	led0		450 nm	LED blau
; 	PB1:	led1		950 nm	LD274-3 Osram
; 	PB2:	led2		1,6 m	led10#16-1
; 	PB3:	led3		2,3 m	led10#23-1
; 
; 	PB4: STOP-Signaleingang invers zu Quinto GLC403, BU X15:
;				PB4=low: 	Strom aus (/STROM)
;	
;	(Wird PB4 nicht beschaltet, funktioniert wegen eines Pullup an PB4 
;	dennoch der Plasmasimulator.)
; 
; 
; Format der Zeitfunktionen im File *.inc (am Ende des Programms):
; - Testdaten als Vektor db_daten:	
; 	pro Step werden vier Amplitudenwerte a vier Bit = 2 Byte
; 	bergeben.
; 	Volumen maximal 256 Steps * 2 Byte = 512 Byte 
; 	Vorsicht: ATtiny13-Flash hat nur 512 Worte a 16 bit
; 	Testdaten: 256 Worte a 2 Byte = 512 Byte
; 	Datenreihenfolge: PB0 PB1 - PB2 PB3		(2 unsigned Bytes)
; 
; Filterung PWM zu analog: je 10F Keramik parallel zur LED,
;	(besser wre ein zweistufiges Filter vor der Endstufe)
; 
; PWM:
; 	Pulsweitenmodulation
; 	Puls + Pause					dt = 100 s
; 	Helligkeit regelbar in 			16 Stufen ~ 4 Bit
; 	Grundtakt 						100 s/16 = 6,25 s (160 kHz)
;	Befehlsrate						1 / 9.6 MHz = 104 ns pro Befehl
; 	Verfgbare Maschinenzyklen 		9,6 MHz/160 kHz = 60 pro Interrupt
; 
; Puls/Pause:
; 	Pulsdauer whlbar 0,5...6,4 ms  ~ typ. 5...64 steps = 0x00...0x50
; 	Rest zu 255 ist Pause 
; 
; Algorithmus:
; 	Zeitschritt:
; 	- Entnahme von vier Amplitutuden aus .db_daten in *.inc-Datei
; 	- bergabe der Amplituden an die PWM-Register
; 	PWM-Schritt:
; 	- Timerinterrupt alle 6,25 s
; 	- PWM-Zyklus von 0...15
; 	- Interrupts zhlen, um Zeitschritt zu erhalten
; 	- Abfrage der vier Amplituden 
; 	- Portabschaltung der getroffenen Pins bei Match
; 	Ende der .dB_daten: von vorn beginnen
; 
; PWM-Werte:
; 	db_daten-Byte 1
; 		Nibble high: 	led1
; 		Nibble low: 	led0
; 	db_daten-Byte 2
; 		Nibble high: 	led3
; 		Nibble low: 	led2
; 	.db-Datenfolge PB0_PB1, PB2_PB3,\
; 
; AVR Fuses:
;	CKDIV8 		aus
;	BODLEVEL 	4,3V
;	SUT_CKSEL	Interner RC-Osz. 9,6 MHz, 14 Ck + 64 ms
;		HIGH	0xF9
; 		LOW		0x7A
;
; Timing
; ------
;
; Zeitquantelung des Gesamtzyklus (256 Samples)
;
;  |--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|   t
;  0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f  0 [ZL]
;  0 1,6  ...								...  25,6  ms
; 
; 	ein Zeitschritt ~ 100 s, 
; 	dieser ist wiederum in 16 PWM-Schritte unterteilt
; 	Zyklus 8 bit = 256 Schritte ~ 256 * 100 s = 25,6 ms
; 
; Zeitschema des einzelnen PWM-Zyklus (4 Bit = 16 Stufen)
; 
;  |--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|   t
;  0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f  0 [pwm]
;  0  6  12 18  ...								  100  s
; 
; 	LEDs bleiben 0...15 Zeitschritte lang an (PWM), Rest bis 15 aus
;
;
; Grundtakte:
;	timesteps 	Basic cycle 	25,6 ms	~ 256 Schritte ~ 256 * 100 s
;	pwmsteps	PWM cycle		100 s 	~ 16 * 6,25 s
;	clocks		PWM clock		6,25 s	~ 60 * 104,166 ns 
;	interner clock 				104,166 ns ~ 9,6 MHz (RC-Oszillator)
;
.equ timesteps 	= 0xff		; Anzahl der Samples (max. 256)
							; mu kleiner/gleich Lnge der db_daten sein!
.equ pwmsteps 	= 0x0f		; 16-stufige PWM-Helligkeit

.equ clocks		= 0x3c		; Ziel: 0x3c = dez.60 ~ 6,25s
							; 6,25s * 16 = 0,1ms
							; 0,1ms mit Logikanalysator prfen
							; Timer-Interrupt alle 60 clocks
							; dh. alle 6,25s bei 9,6MHz

.equ prescal 	= 1		; 1: ohne Prescaler 9,6MHz ~ 0,104166 s pro Befehl
;.equ prescal 	= 5		; 5: fr langsame Tests /1024 = 9375 Hz
; Set prescaler TCCR0B 2:0 = CS0 2:0
		; 	/1: 	001		1
		; 	/8: 	010		2
		; 	/64: 	011		3
		; 	/256: 	100		4
		; 	/1024: 	101		5

; Tests for Watch:
;	pwm		PWM-zhler		(00-0f)
;	sample	Zeilenzhler	(00-ff)
;	led0	LED-Wert PB0	(00-0f)
;	led1	PB1
;	led2	PB2
;	led3	PB3
;	zl		Datenpointer	(00-ff)


;############### Namen ###################
Anfang:

; Register dynamisch:
.def 	A1	= r16		; Arbeit
.def	A2	= r17		; Arbeit
.def	A3	= r18		; Arbeit
.def	A4	= r19		; Arbeit


; Register statisch je ber 16 PWM-Steps:
.def	led0 	= r20	; Helligkeitswerte	 450 nm
.def	led1 	= r21	; Helligkeitswerte	 940 nm
.def	led2 	= r22	; Helligkeitswerte	1500 nm
.def	led3 	= r23	; Helligkeitswerte	2300 nm

; statisch:
.def	pwm 	= r24	; PWM-Zhler 0...15
.def 	sample	= r25	; Zeitschrittzhler 0...256


/* 
; aus tn13def.inc wird benutzt:
.def	ZL	= r30		; zur Adressierung der Daten-Bytes
.def	ZH	= r31
*/

; ############ Inits #########################################


rjmp Reset

; Interrupt-Adresse reservieren (Compiler-Anweisung)
;############### Adressen ################

;	0x0000 rjmp RESET 		; Reset Handler
;	0x0001 rjmp EXT_INT0 	; IRQ0 Handler
;	0x0002 rjmp PCINT0 		; PCINT0 Handler
;	0x0003 rjmp TIM0_OVF 	; Timer0 Overflow
;	0x0004 rjmp EE_RDY 		; EEPROM Ready Handler
;	0x0005 rjmp ANA_COMP 	; Analog Comparator Handler
;	0x0006 rjmp TIM0_COMPA 	; Timer0 Compare A
;	0x0007 rjmp TIM0_COMPB 	; Timer0 Compare B
;	0x0008 rjmp WATCHDOG 	; Watchdog Interrupt Handler
;	0x0009 rjmp ADC 		; ADC Conversion Handler
							; siehe Tab.9-1, S.44

; Pegel-Wechsel an PB4 beobachten, S.47
.org 0x0002 rjmp PinChangeInterrupt	; bei PB4-nderung neu starten

; Timer/Counter Compare Match A, S.44
.org 0x0006	rjmp TimerInterrupt   ; gehe zum Interupt-Behandlungsprogramm

.cseg	; Codesegment folgt

Reset:
; Stack setzen:
	ldi A1,low(RAMEND)		; Set Stack Pointer to top of RAM
	out SPL,A1 				; (SPH gibt es beim ATtiny13 nicht)


rjmp Main		; springe zum Hauptprogramm

;#############################################################
;####################### Unterprogramme ######################
;#############################################################


InitTimerCompA:		; Unterbrechung alle 100 s/16 = 6,25 s (160 kHz)
						; 9,6 MHz/160 kHz = 60 pro Interrupt
						; Timer-Interrupt Init compare match A
						; siehe S.60, 11.5 
						; Grundtakterzeugung: alle 6,25s ~ 160kHz

	; Vorteiler einstellen, S.73, Tab.11-9
	ldi A1,prescal		; Timer-Prescaler laden (CS0)
	out TCCR0B,A1		; 

	; Fast PWM mode, S.64, 11.7.3  
	; WGM2:0 = 7 dazu 
	
	; WGM01:0 = 11 setzen
	ldi A1,0x03			; 0b0000_0011, Seite 72, Tab. 11-8
	out TCCR0A,A1

	; WGM02 = 1 setzen
	in  A2,TCCR0B
	ori A2,0x08			; Einsen bernehmen
	out TCCR0B,A2

	; Timer Interrupt freigeben:
	ldi A1,4			; 0b0000_0100 OCIE0A wird gesetzt, 
	out TIMSK0,A1	  	; Timer Interrupt freigeben S.64, S.74

	; Zeitdauer fr Interrupt in Timer-Register OCR0A
	ldi A3,clocks		; nach Anzahl "clocks" Interrupt auslsen
	out OCR0A,A3		; Timer-Register A laden
ret


Init_Ports:
	; Datenrichtung S.51	0: input, 1: output
		ldi A1,0b_0000_1111	; PB7...0
		out ddrb,A1			; PB0...3 auf Ausgabe, PB4...7 Input

	; Ausgabe S.51
		ldi A1,0b_1111_1111	; PB7...0
		out portb,A1		; PB0...3 LEDs leuchten bei high
							; PB4...7 auf pull-up (1)
ret


InitPinChangeInter:	; Pin change Interrupt an PB4 aktivieren
					; S.47, S.156
	; Portbit 4 aktivieren		; PCMSK |= (1<<PINB4);
		; sbi PCMSK,PCINT4		; get fei aah!
		ldi A4,0b0001_0000
		out PCMSK,A4			; load pin change enable mask

	; Flanke einstellen
		; MCUCR = (1<<ISC01) | (1<<ISC00);
		;	
		;	ISC01	ISC00
		;	0		0		low level
		;	0		1		pin change
		;	1		0		falling edge
		;	1		1		rising edge
		; 	
		; sbi MCUCR,ISC01	; get fei net
		; sbi MCUCR,ISC00	; get aah net
		;ldi A4,0b0000_0010	; falling edge
		ldi A4,0b0000_0000	; low level
		out MCUCR,A4

	; Interrupts freigeben
		; sbi GIMSK,INT0	; get net
		; sbi GIMSK,PCIE	; get net
		;ldi A4,0b0110_0000	; PCIE and INT0 - get net
		ldi A4,0b0010_0000	; PCIE - pin change interrupt enable
		out GIMSK,A4		; to general interrupt mask register
ret


InitDataptr:			; Datenpointer fr *.inc File initialisieren
					; Adresse des Labels db_daten auf ZH/ZL laden

	ldi ZH,high(db_daten*2)		; Adresse vom Label db_daten nach Z holen
	ldi ZL,low(db_daten*2)		; Flash ist in 2-Byte organisiert (*2)

ret


GetValues:	; Amplitudenwerte 1...4 aus *.inc holen, dazu
			; zwei Byte aus db_daten holen, Adresse in ZH,ZL

	lpm led0,Z+	; erstes Byte aus Z-Reg. laden, postincrement
	lpm led2,Z+	; zweites Byte aus Z-Reg. laden, postincrement
				; LOAD PROGRAM MEMORY (in Wortorganisation *2)
				; LPM loads one byte pointed to by the Z register
				; Z zhlt jedes Byte weiter! 
				; Da zwei Bytes pro Sample geholt werden, 
				; erhht sich Z um zwei pro Step
				; deshalb Doppelzhlung mit sample

	rcall Byte2Nibble	; auf led0...led3 konvertieren
ret


Byte2Nibble:	; aus zwei Byte vier Nibbles in Bytes machen
				; in: 	Bytes 		led0 	   led2
				; out: 	Bytes 	led0, led1, led2, led3
				; led0, led2 msb
				; led1, led3 lsb
				; Daten led0 led1, led2 led3,\
	
	mov led1,led0		; copy led0 to led1
	swap led0			; swap high and low nibble
	andi led0, 0x0f		; delete high nibble
	andi led1, 0x0f		; delete high nibble

	mov led3,led2		; copy led2 to led3
	swap led2			; swap high and low nibble
	andi led2, 0x0f		; delete high nibble
	andi led3, 0x0f		; delete high nibble
ret


LEDs_an:				; LEDs PB0-3 zu Beginn des PWM-Zyklus einschalten
	in A1,portb
	ori A1,0x0f			; set low nibble, PB0-3 LEDs
	out portb,A1
ret


LEDs_aus:				; z.B. whrend einer STOP-Pause
	in A1,portb
	andi A1,0xf0		; reset low nibble, PB0-3, LEDs
	out portb,A1
ret


LEDs_toggle:			; nur fr Tests
	com led0			; Negation
	out portb,led0		; LEDs umschalten	
ret



;**********************************************
; 			Interrupt-Behandlung
;**********************************************

; initial: pwm=0, sample=0, Leds an (H)

TimerInterrupt:			; Timer Interrupt bei compare match A
						; dies ist das eigentliche Hauptprogramm,
						; es wird alle 6,25 s ausgefhrt
						; 6,25s * 9,6 MHz = 60 Befehle stehen zur Verfgung
	sei					; Timer-Interrupt sofort wieder starten, 
						; damit nur 60 Takte folgen
						; (sonst wrde dies erst der reti-Befehl tun)
						; Vorsicht: nchstes Interrupt kann dann reinhacken!

	; Pause zwischen ein- und aus (Eins-Spike) minimieren:
		cpi pwm,0			; fngt gerade ein neuer pwm zyklus an?
		brne timtr			; nein
		rcall LEDs_an		; alle LEDs einschalten (place breakpoint here!)
		timtr:

	; PWM erzeugen:

		cp led0, pwm		; Helligkeitswerte mit PWM Zhlerstand vergleichen
			brne tintr1		; branch, wenn led ungleich pwm
			cbi portb,PB0	; wenn nicht, LED ausschalten
			tintr1:			; auf led0...4 liegen nur rechte Nibbles a 4 Bit

		cp led1, pwm
			brne tintr2
			cbi portb,PB1
			tintr2:

		cp led2, pwm
			brne tintr3
			cbi portb,PB2
			tintr3:

		cp led3, pwm
			brne tintr4
			cbi portb,PB3
			tintr4:
									
		inc pwm				; PWM-Zhler inkrementieren

	; Durchlauf des PWM Zyklus prfen:
		cpi pwm,pwmsteps	; PWM-Zhlerstand vergleichen
		brlo timrdy			; branch if lower

		rcall GetValues		; nchste Zeile fr LEDs holen
		inc sample			; Timestep-Zhler erhhen
		ldi pwm,0			; PWM-Zhler neu laden

	; Ablauf des Timestep-Zhlers prfen (ganz von vorn):
		cpi sample,timesteps	; Zyklus prfen
		brlo timrdy			; branch if lower
		ldi sample,0		; Timestep-Zhler neu laden
		rcall InitDataptr	; Datenpointer rcksetzen auf Startadresse
		rcall GetValues		; neue Werte fr LEDs holen

		timrdy:

reti	; Warte auf das nchste Interrupt


PinChangeInterrupt:		; Restart nach Ende des STOP-Puls vom Spektralregler
	; Schwei-Strom aus bei PB4 = low
	; nach ext.Interrupt auf PB4, fallende Flanke: 
	; Interrupt disable:
	cli

	; alles rcksetzen
		ldi pwm,0			; PWM-Zhler rcksetzen
		ldi sample,0		; Zeitschrittzhler rcksetzen
		ldi ZH,0			; Z Register auf Null setzen,
		ldi ZL,0			; Z luft doppelt so schnell, wie sample
		ldi sample,0		; Timestep-Zhler neu laden
		rcall LEDs_aus		; LEDs abschalten

	; solange STOP (PB4 = null), hier verweilen
		pinhalt:
		sbis pinb,PB4		; skip if bit is set
		rjmp pinhalt		; bit not set, wait
		nop
		
	; Interrupt enable, neue Daten holen	
		rcall InitDataptr	; Datenpointer rcksetzen auf Startadresse
		rcall GetValues		; neue Werte fr LEDs holen
		rcall InitPinChangeInter
		rcall InitTimerCompA
	sei
reti


;#############################################################
;######################### MAIN ##############################
;#############################################################


Main:

; Inits
	rcall Init_Ports

	ldi pwm,0			; PWM-Zhler initialisieren
	ldi sample,0		; Zeitschrittzhler-Init

	ldi ZH,0			; Z-Register auf Null setzen,
	ldi ZL,0			; Z luft doppelt so schnell, wie sample

	rcall InitDataptr	; Daten init
	rcall GetValues		; erstes Doppelbyte holen
	rcall InitPinChangeInter
	rcall InitTimerCompA
	sei   					; Interrupt freigeben

mainloop:			; auf Interrupts warten
rjmp mainloop		; Rest findet im Timer-Interrupt statt


;#############################################################
;######################### DATEN #############################
;#############################################################

.cseg

;.include "saegezahn256.inc"		; Zeitfunktion fr Test
.include "A664-0398_2010_04.inc"	; Schweipuls einbinden

; Ende
